@@ -28,6 +28,9 @@ from .platform import (
     SchemaVersion,
     Status,
     TagNameType,
+    Team,
+    TeamAffiliation,
+    TeamMember,
     body_maxlen,
     summary_maxlen,
     title_maxlen,
@@ -48,12 +51,14 @@ class Command(SchemaVersion, DyffSchemaBaseModel):
 
     command: Literal[
         "CreateChallengeTask",
+        "CreateChallengeTeam",
         "CreateEntity",
         "EditChallengeContent",
         "EditChallengeTaskRules",
         "EditEntityDocumentation",
         "EditEntityLabels",
         "EditFamilyMembers",
+        "EditTeam",
         "ForgetEntity",
         "RestoreEntity",
         "UpdateEntityStatus",
@@ -101,6 +106,31 @@ class CreateChallengeTask(Command):
 # ----------------------------------------------------------------------------
 
 
+class CreateChallengeTeamAttributes(DyffSchemaBaseModel):
+    """Attributes for the CreateChallengeTeam command."""
+
+    team: Team = pydantic.Field(description="The team to create.")
+
+
+class CreateChallengeTeamData(EntityIdentifier):
+    """Payload data for the CreateChallengeTeam command."""
+
+    attributes: CreateChallengeTeamAttributes = pydantic.Field(
+        description="The command attributes"
+    )
+
+
+class CreateChallengeTeam(Command):
+    """Create a new team participating in an existing challenge."""
+
+    command: Literal["CreateChallengeTeam"] = "CreateChallengeTeam"
+
+    data: CreateChallengeTeamData = pydantic.Field(description="The command data.")
+
+
+# ----------------------------------------------------------------------------
+
+
 class ChallengeContentPagePatch(JsonMergePatchSemantics):
     """Same properties as ChallengeContentPage.
 
@@ -188,9 +218,6 @@ class ChallengeTaskRulesExecutionEnvironmentPatch(JsonMergePatchSemantics):
     choices: dict[str, Optional[ChallengeTaskExecutionEnvironment]] = pydantic.Field(
         default_factory=dict, description="Execution environment choices."
     )
-    default: Optional[str] = pydantic.Field(
-        default=None, description="The default execution environment."
-    )
 
 
 class ChallengeTaskRulesSchedulePatch(JsonMergePatchSemantics):
@@ -227,20 +254,15 @@ class ChallengeTaskRulesSchedulePatch(JsonMergePatchSemantics):
 
 
 class ChallengeTaskRulesPatch(DyffSchemaBaseModel):
-    """Same properties as ChallengeTaskRules, but assigning None to a field is
-    interpreted as a command to delete that field.
+    """Edits to make to the different rules categories."""
 
-    Fields that are not assigned explicitly remain unchanged.
-    """
-
-    executionEnvironment: Optional[ChallengeTaskRulesExecutionEnvironmentPatch] = (
-        pydantic.Field(
-            default=None,
-            description="Patch for the .rules.executionEnvironments field.",
-        )
+    executionEnvironment: ChallengeTaskRulesExecutionEnvironmentPatch = pydantic.Field(
+        default_factory=ChallengeTaskRulesExecutionEnvironmentPatch,
+        description="Patch for the .rules.executionEnvironments field.",
     )
-    schedule: Optional[ChallengeTaskRulesSchedulePatch] = pydantic.Field(
-        default=None, description="Patch for the .rules.schedule field."
+    schedule: ChallengeTaskRulesSchedulePatch = pydantic.Field(
+        default_factory=ChallengeTaskRulesSchedulePatch,
+        description="Patch for the .rules.schedule field.",
     )
 
     @field_serializer("executionEnvironment")
@@ -260,10 +282,18 @@ class ChallengeTaskRulesPatch(DyffSchemaBaseModel):
         return schedule.model_dump(mode=_info.mode)
 
 
+class EditChallengeTaskRulesPatch(DyffSchemaBaseModel):
+    """A Json Merge Patch for the rules of one task."""
+
+    rules: ChallengeTaskRulesPatch = pydantic.Field(
+        description="Edits to make to the task rules."
+    )
+
+
 class EditChallengeTaskRulesAttributes(DyffSchemaBaseModel):
     """Attributes for the EditChallengeTaskRules command."""
 
-    rules: ChallengeTaskRulesPatch = pydantic.Field(
+    tasks: dict[str, EditChallengeTaskRulesPatch] = pydantic.Field(
         description="Edits to make to the task rules."
     )
 
@@ -433,6 +463,40 @@ class EditFamilyMembers(Command):
 # ----------------------------------------------------------------------------
 
 
+class EditTeamAttributes(JsonMergePatchSemantics):
+    """Attributes for the EditFamilyMembers command."""
+
+    members: dict[str, Optional[TeamMember]] = pydantic.Field(
+        default_factory=dict, description="The members of this team"
+    )
+    affiliations: dict[str, Optional[TeamAffiliation]] = pydantic.Field(
+        default_factory=dict,
+        description="The affiliations of the team. Team members state their"
+        " affiliations by referencing these entries by their keys.",
+    )
+
+
+class EditTeamData(EntityIdentifier):
+    """Payload data for the EditFamilyMembers command."""
+
+    attributes: EditTeamAttributes = pydantic.Field(
+        description="The command attributes"
+    )
+
+    @field_serializer("attributes")
+    def _serialize_attributes(self, attributes: EditTeamAttributes, _info):
+        return attributes.model_dump(mode=_info.mode)
+
+
+class EditTeam(Command):
+    command: Literal["EditTeam"] = "EditTeam"
+
+    data: EditTeamData = pydantic.Field(description="The edit data.")
+
+
+# ----------------------------------------------------------------------------
+
+
 class ForgetEntity(Command):
     """Forget (permanently delete) an entity."""
 
@@ -516,11 +580,14 @@ class UpdateEntityStatus(Command):
 
 DyffCommandType = Union[
     CreateChallengeTask,
+    CreateChallengeTeam,
     CreateEntity,
     EditChallengeContent,
+    EditChallengeTaskRules,
     EditEntityDocumentation,
     EditEntityLabels,
     EditFamilyMembers,
+    EditTeam,
     ForgetEntity,
     RestoreEntity,
     UpdateEntityStatus,
@@ -537,15 +604,19 @@ __all__ = [
     "CreateChallengeTask",
     "CreateChallengeTaskAttributes",
     "CreateChallengeTaskData",
+    "CreateChallengeTeam",
+    "CreateChallengeTeamAttributes",
+    "CreateChallengeTeamData",
     "CreateEntity",
     "DyffCommandType",
     "EditChallengeContent",
     "EditChallengeContentAttributes",
     "EditChallengeContentData",
     "EditChallengeContentPatch",
+    "EditChallengeTaskRules",
     "EditChallengeTaskRulesAttributes",
     "EditChallengeTaskRulesData",
-    "EditChallengeTaskRules",
+    "EditChallengeTaskRulesPatch",
     "EditEntityDocumentation",
     "EditEntityDocumentationAttributes",
     "EditEntityDocumentationData",
@@ -556,6 +627,9 @@ __all__ = [
     "EditFamilyMembers",
     "EditFamilyMembersAttributes",
     "EditFamilyMembersData",
+    "EditTeam",
+    "EditTeamAttributes",
+    "EditTeamData",
     "EntityIdentifier",
     "FamilyIdentifier",
     "ForgetEntity",
