@@ -486,6 +486,13 @@ class ScoreQueryRequest(DyffRequestDefaultValidators):
     model: Optional[str] = pydantic.Field(default=None)
 
 
+class SubmissionQueryRequest(DyffEntityQueryRequest):
+    challenge: Optional[str] = pydantic.Field(default=None)
+    task: Optional[str] = pydantic.Field(default=None)
+    team: Optional[str] = pydantic.Field(default=None)
+    pipelineRun: Optional[str] = pydantic.Field(default=None)
+
+
 class TeamQueryRequest(DyffEntityQueryRequest):
     challenge: Optional[str] = pydantic.Field(default=None)
 
@@ -542,6 +549,7 @@ __all__ = [
     "SafetyCaseQueryRequest",
     "ScoreQueryRequest",
     "SubmissionCreateRequest",
+    "SubmissionQueryRequest",
     "TeamEditRequest",
     "TeamQueryRequest",
     "UseCaseQueryRequest",
